#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# D-Case Weaver 1.1.1
#
# Copyright (c) 2012-2014 JST DEOS R&D Center
#
# Licensed under The MIT License
#

import base64
import cgi
from cmislib import CmisClient
import json
import os
import re
import StringIO
import syslog
import traceback

from cmislib.exceptions import CmisException, PermissionDeniedException, ObjectNotFoundException
from urllib2 import URLError

Status = {
    200: "200 OK",
    201: "201 Created",
    400: "400 Bad Request",
    401: "401 Unauthorized",
    404: "404 Not Found",
    500: "500 Internal Server Error",
    550: "550 Permission denied"
};


def debug (msg):
    if isinstance(msg, unicode):
        msg = msg.encode("utf-8")
    elif not isinstance(msg, str):
        msg = str(msg)

    syslog.syslog(syslog.LOG_DEBUG, msg)

class AlfException(Exception):

    """
    Common base class for all exceptions.
    """

    def __init__(self, message=None):
        Exception.__init__(self, "Error %s" % message)
        self.message = message

class UnauthorizedException(AlfException):
    pass

class BadRequestException(AlfException):
    pass

class SQLQuery (object):
    def __init__ (self, repo):
        self.repo = repo;
        self.conditions = []

    def addCondition (self, condition):
        self.conditions.append(condition)

    def query (self):
        query = "SELECT * from cmis:document "
        for i in range(len(self.conditions)):
            if i == 0:
                query += "WHERE " + self.conditions[i].queryString()
            else:
                query += " AND " + self.conditions[i].queryString()

        return self.repo.query(query, searchAllVersions="true")

class SQLCondition (object):
    def queryString (self):
        pass

class SQLConditionName (SQLCondition):
    def __init__ (self, name, op):
        self.name = name
        self.op = {"=": "like", "<>": "not like"}[op]

    def queryString (self):
        return "cmis:name %s '%%%s%%'" % (self.op, self.name)

class SQLConditionCreatedBy (SQLCondition):
    def __init__ (self, user, op):
        self.user = user
        self.op = {"=": "like", "<>": "not like"}[op]

    def queryString (self):
        return "cmis:createdBy %s '%%%s%%'" % (self.op, self.user)

class SQLConditionCreationDate (SQLCondition):
    def __init__ (self, date, op):
        self.date = date
        self.op = op

    def queryString (self):
        return "cmis:creationDate %s '%s'" % (self.op, self.date)

class SQLConditionModifiedBy (SQLCondition):
    def __init__ (self, user, op):
        self.user = user
        self.op = {"=": "like", "<>": "not like"}[op]

    def queryString (self):
        return "cmis:lastModifiedBy %s '%%%s%%'" % (self.op, self.user)

class SQLConditionModificationDate (SQLCondition):
    def __init__ (self, date, op):
        self.date = date
        self.op = op

    def queryString (self):
        return "cmis:lastModificationDate %s '%s'" % (self.op, self.date)

class SQLConditionComment (SQLCondition):
    def __init__ (self, comment, op):
        self.comment = comment
        self.op = {"=": "like", "<>": "not like"}[op]

    def queryString (self):
        return "cmis:checkinComment %s '%%%s%%'" % (self.op, self.comment)

class Condition (object):
    def check (self):
        pass

class ConditionSet (Condition):
    def __init__ (self):
        self.conditions = []

    def check (self, doc):
        for c in self.conditions:
            if not c.check(doc):
                return False
        return True

    def add (self, condition):
        self.conditions.append(condition)

class ConditionDirectoryPath (Condition):
    def __init__ (self, path, repo):
        self.verify(path, repo)
        self.path = os.path.normpath(path) + "/"

    def verify (self, path, repo):
        errmsg = ("Directory is not found: %s" % path)
        try:
            obj = repo.getObjectByPath(path)
        except:
            raise BadRequestException(errmsg)

        if obj.properties["cmis:baseTypeId"] != "cmis:folder":
            raise BadRequestException(errmsg)

    def check (self, doc):
        for path in doc.getPaths():
            if re.match(self.path, path.encode("utf-8")) is not None:
                return True
        return False

class ConditionComment (Condition):
    def __init__ (self, comment, op):
        self.comment = comment
        self.op = op

    def check (self, doc):
        resultSet = doc.getAllVersions()
        for result in resultSet.getResults():
            comment = result.properties.get("cmis:checkinComment")
            if comment is not None:
                if re.match(self.comment, comment.encode("utf-8")):
                    return True
        return False

class Alfresco(object):

    def __init__(self):
        self.client = None

    def __call__(self, environ, start_response):

        headers = [("Content-type", "text/plain"), ('Cache-Control','no-cache')]
        status = output = Status[400]

        try:
            self.environ = environ

            pathinfo = environ["PATH_INFO"]

            if len(pathinfo) == 0:
                status = output = Status[200]
                return

            if pathinfo == "/login/":
                status, headers, output = self.login()
                return

            if pathinfo == "/logout/":
                status, headers, output = self.logout()
                return


            q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))
            if "ba" not in q \
                    or "repositoryUrl" not in q:
                raise UnauthorizedException("Unauthorized Exception")

            user_pass = base64.b64decode(q["ba"]).split(":")

            self.client = CmisClient(
                q["repositoryUrl"], user_pass[0], user_pass[1])


            m = re.match("^/repository/(.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                status, headers, output = self.repository(repoId)
                return

            m = re.match("^/metadata/([^/]+)/p(/.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                path = m.group(2)
                status, headers, output = self.metadata(repoId, path=path)
                return

            m = re.match("^/metadata/([^/]+)/i/([^/]+)/(.+)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                fileId = "%s/%s" % (m.group(2).replace(":", "://"), m.group(3))
                status, headers, output = self.metadata(repoId, fileId=fileId)
                return

            m = re.match("^/parents/([^/]+)/p(/.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                path = m.group(2)
                status, headers, output = self.parents(repoId, path=path)
                return

            m = re.match("^/parents/([^/]+)/i/([^/]+)/(.+)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                fileId = "%s/%s" % (m.group(2).replace(":", "://"), m.group(3))
                status, headers, output = self.parents(repoId, fileId=fileId)
                return

            m = re.match("^/children/([^/]+)/p(/.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                path = m.group(2)
                status, headers, output = self.children(repoId, path=path)
                return

            m = re.match("^/children/([^/]+)/i/([^/]+)/(.+)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                fileId = "%s/%s" % (m.group(2).replace(":", "://"), m.group(3))
                status, headers, output = self.children(repoId, fileId=fileId)
                return

            m = re.match("^/files/([^/]+)/p(/.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                path = m.group(2)
                status, headers, output = self.files(repoId, path=path)
                return

            m = re.match("^/files/([^/]+)/i/([^/]+)/(.+)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                fileId = "%s/%s" % (m.group(2).replace(":", "://"), m.group(3))
                status, headers, output = self.files(repoId, fileId=fileId)
                return

            m = re.match("^/mkdir/([^/]+)/p(/.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                path = m.group(2)
                status, headers, output = self.mkdir(repoId, path=path)
                return

            m = re.match("^/mkdir/([^/]+)/i/([^/]+)/(.+)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                fileId = "%s/%s" % (m.group(2).replace(":", "://"), m.group(3))
                status, headers, output = self.mkdir(repoId, fileId=fileId)
                return

            m = re.match("^/search/([^/]+)$", pathinfo)
            if m is not None:
                if "conditions" not in q:
                    raise BadRequestException("Bad Request")
                conditions = json.loads(q["conditions"])
                repoId = m.group(1)
                status, headers, output = self.search(repoId, conditions)
                return

            m = re.match("^/approve/([^/]+)/i/([^/]+)/(.+)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                fileId = "%s/%s" % (m.group(2).replace(":", "://"), m.group(3))
                status, headers, output = self.approval(repoId, fileId=fileId)
                return

            m = re.match("^/approve/([^/]+)/p(/.*)$", pathinfo)
            if m is not None:
                repoId = m.group(1)
                path = m.group(2)
                status, headers, output = self.approval(repoId, path=path)
                return

            raise BadRequestException("Bad command")

        except URLError as e:
            status = Status[404]
            output = "%s %s" % (str(type(e)), str(e))

        except PermissionDeniedException as e:
            status = Status[550]
            output = "%s %s" % (str(type(e)), str(e))

        except ObjectNotFoundException as e:
            status = Status[404]
            output = "%s %s" % (str(type(e)), str(e))

        except CmisException as e:
            status = Status[500]
            output = "%s %s" % (str(type(e)), str(e))

        except BadRequestException as e:
            status = Status[400]
            output = "%s %s" % (str(type(e)), str(e))

        except UnauthorizedException as e:
            status = Status[401]
            output = "%s %s" % (str(type(e)), str(e))

        except Exception as e:
            status = Status[500]
            output = "%s %s" % (str(type(e)), str(e))

        except:
            status = output = Status[500]

        finally:
            start_response(status, headers)
            return output


    def login(self):
        q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))
        if "ba" not in q \
                or "repositoryUrl" not in q:
            raise UnauthorizedException("Unauthorized Exception")

        user_pass = base64.b64decode(q["ba"]).split(":")

        self.client = CmisClient(q["repositoryUrl"], user_pass[0], user_pass[1])
        try:
            self.client.getRepositories()

        except PermissionDeniedException:
            raise UnauthorizedException("Unauthorized Exception")

        headers = [("Content-type", "text/plain"), ('Cache-Control','no-cache')]
        return Status[200], headers, "Authentication succeeded."


    def logout(self):
        headers = [("Content-type", "text/plain")]

        self.client = None

        return Status[200], headers, "Logout"


    def repository(self, repoId):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]

        method = self.environ["REQUEST_METHOD"]
        if method == "GET":

            if repoId == "":
                repos = self.client.getRepositories()
                return Status[200], headers, json.dumps(repos)
            else:
                pass

        return Status[400], headers, Status[400]


    def metadata (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]

        q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))

        repo = self.client.getRepository(repoId)

        obj = None

        path = None

        if "path" in kwargs:
            path = kwargs["path"]
            obj = repo.getObjectByPath(path)
        elif "fileId" in kwargs:
            obj = repo.getObject(kwargs["fileId"])
            path = obj.getPaths()[0]
        else:
            raise BadRequestException("Bad Request: url")

        if obj.properties["cmis:baseTypeId"] == "cmis:folder":
            ret = self.normalizeProperties(obj.properties)
            ret["path"] = path
            children = []
            for child in obj.getChildren():
                prop = self.normalizeProperties(child.properties)
                prop["path"] = child.getPaths()[0]
                children.append(prop)
            ret["children"] = children;

        elif obj.properties["cmis:baseTypeId"] == "cmis:document":
            resultSet = obj.getAllVersions()
            results = resultSet.getResults()

            if q.get("allVersions", "false") == "true":
                ret = []
                for result in results:
                    prop = self.normalizeProperties(result.properties)
                    prop["path"] = path
                    ret.append(prop)

            else:
                prop = None

                version = q.get("version", "latest")

                for result in results:
                    if result.properties["cmis:isLatestVersion"]:
                        prop = result.properties
                        prop["path"] = path
                        break

                if version == "latest":
                    pass
                else:
                    for result in results:
                        if result.properties["cmis:versionLabel"] == version:
                            prop = result.properties
                            prop["path"] = path
                            break

                ret = self.normalizeProperties(prop)

        return Status[200], headers, json.dumps(ret)


    def parents (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]

        repo = self.client.getRepository(repoId)

        obj = None
        path = None

        if "path" in kwargs:
            path = kwargs["path"]
            obj = repo.getObjectByPath(path)
        elif "fileId" in kwargs:
            obj = repo.getObject(kwargs["fileId"])
            path = obj.getPaths()[0]
        else:
            raise BadRequestException("Bad Request: url")

        ret = []

        parents = obj.getObjectParents()
        for parent in parents:
            prop = self.normalizeProperties(parent.properties)
            prop["path"] = parent.getPaths()[0]
            ret.append(prop)

        return Status[200], headers, json.dumps(ret)


    def children (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]

        repo = self.client.getRepository(repoId)

        obj = None
        path = None

        if "path" in kwargs:
            path = kwargs["path"]
            obj = repo.getObjectByPath(path)
        elif "fileId" in kwargs:
            obj = repo.getObject(kwargs["fileId"])
            path = obj.getPaths()[0]
        else:
            raise BadRequestException("Bad Request: url")

        ret = []

        if obj.properties["cmis:baseTypeId"] == "cmis:folder":
            children = obj.getChildren()
            for child in children:
                if child.properties["cmis:baseTypeId"] == "cmis:folder":
                    prop = self.normalizeProperties(child.properties)
                    prop["path"] = child.getPaths()[0]
                    ret.append(prop)
                elif child.properties["cmis:baseTypeId"] == "cmis:document":
                    resultSet = child.getAllVersions()
                    results = resultSet.getResults()
                    for result in results:
                        if result.properties["cmis:isLatestVersion"]:
                            break
                    prop = self.normalizeProperties(result.properties)
                    prop["path"] = child.getPaths()[0]
                    ret.append(prop)

        return Status[200], headers, json.dumps(ret)


    def files(self, repoId, **kwargs):

        method = self.environ["REQUEST_METHOD"]

        if method == "GET":
            repo = self.client.getRepository(repoId)

            q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))

            obj = None

            if "path" in kwargs:
                obj = repo.getObjectByPath(kwargs["path"])
            elif "fileId" in kwargs:
                obj = repo.getObject(kwargs["fileId"])
            else:
                raise BadRequestException("Bad Request: url")

            if obj.properties["cmis:baseTypeId"] == "cmis:folder":
                raise BadRequestException("Bad Request: path")

            resultSet = obj.getAllVersions()
            results = resultSet.getResults()

            version = q.get("version", "latest")

            for result in results:
                if result.properties["cmis:isLatestVersion"]:
                    obj = result
                    break

            if version == "latest":
                pass
            else:
                for result in results:
                    if result.properties["cmis:versionLabel"] == version:
                        obj = result
                        break

            mime = obj.properties["cmis:contentStreamMimeType"].encode("utf-8")
            filename = obj.properties["cmis:name"].encode("utf-8")
            headers = [
                ("Content-type", mime),
                ('Cache-Control','no-cache'),
                ("Content-Disposition", "filename=" + filename)]

            return Status[200], headers, obj.getContentStream()

        elif method == "PUT":
            return self.files_put(repoId, **kwargs)

        elif method == "POST":
            return self.files_post(repoId, **kwargs)

        else:
            raise BadRequestException("Bad Request: REQUEST_METHOD")


    def files_put (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]
        status = Status[200]

        doc = None
        path = None

        repo = self.client.getRepository(repoId)

        q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))

        wsgiinput = self.environ['wsgi.input']
        length = int(self.environ.get('CONTENT_LENGTH', 0))

        f = StringIO.StringIO(wsgiinput.read(length))
        try:
            doc = None
            if "path" in kwargs:
                path = kwargs["path"]
                doc = repo.getObjectByPath(path)
            elif "fileId" in kwargs:
                doc = repo.getObject(kwargs["fileId"])
            else:
                raise BadRequestException("Bad Request: url")

            if doc.properties["cmis:baseTypeId"] == "cmis:folder":
                raise BadRequestException("Bad Request: path")

            actions = doc.getAllowableActions()
            if not actions["canCheckOut"]:
                raise PermissionDeniedException()

            pwcDoc = doc.checkout()
            try:
                args = {}

                # bug? Two or more versions progress.
                #if "mimeType" in q:
                #    args["contentType"] = q["mimeType"]

                pwcDoc.setContentStream(f, **args)

                args = {"major": "false"}
                if "comment" in q:
                    args["checkinComment"] = q["comment"]

                doc = pwcDoc.checkin(**args)
            finally:
                if doc.isCheckedOut():
                    pwcDoc.delete()


        finally:
            f.close()

        resultSet = doc.getAllVersions()
        results = resultSet.getResults()
        for result in results:
            if result.properties["cmis:isLatestVersion"]:
                break
        prop = self.normalizeProperties(result.properties)
        prop["path"] = doc.getPaths()[0]

        return status, headers, json.dumps(prop)


    def files_post (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]
        status = Status[200]

        path = None

        repo = self.client.getRepository(repoId)

        q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))
        if "name" not in q:
            raise BadRequestException("Bad Request: name")

        name = q["name"].decode("utf-8")

        wsgiinput = self.environ['wsgi.input']
        length = int(self.environ.get('CONTENT_LENGTH', 0))

        f = StringIO.StringIO(wsgiinput.read(length))
        try:
            folder = None
            if "path" in kwargs:
                path = kwargs["path"]
                folder = repo.getObjectByPath(path)
            elif "fileId" in kwargs:
                folder = repo.getObject(kwargs["fileId"])
            else:
                raise BadRequestException("Bad Request: url")

            if folder.properties["cmis:baseTypeId"] != "cmis:folder":
                raise BadRequestException("Bad Request: path")

            actions = folder.getAllowableActions()
            if not actions["canCreateDocument"]:
                raise PermissionDeniedException()

            args = {}

            args["contentFile"] = f

            # bug? Two or more versions progress.
            #if "mimeType" in q:
            #    args["contentType"] = q["mimeType"]

            doc = folder.createDocument(name, **args)

            status = Status[201]

        finally:
            f.close()

        resultSet = doc.getAllVersions()
        results = resultSet.getResults()
        for result in results:
            if result.properties["cmis:isLatestVersion"]:
                break
        prop = self.normalizeProperties(result.properties)
        prop["path"] = doc.getPaths()[0]

        return status, headers, json.dumps(prop)


    def mkdir (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]
        status = Status[200]

        path = None

        repo = self.client.getRepository(repoId)

        q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))
        if "name" not in q:
            raise BadRequestException("Bad Request: name")

        name = q["name"].decode("utf-8")

        folder = None
        if "path" in kwargs:
            path = kwargs["path"]
            folder = repo.getObjectByPath(path)
        elif "fileId" in kwargs:
            folder = repo.getObject(kwargs["fileId"])
        else:
            raise BadRequestException("Bad Request: url")

        subFolder = folder.createFolder(name)

        prop = self.normalizeProperties(subFolder.properties)
        prop["path"] = subFolder.getPaths()[0]

        return status, headers, json.dumps(prop)


    def search (self, repoId, conditions):
        headers = [("Content-type", "application/json"),
                   ('Cache-Control','no-cache')]
        status = Status[200]

        repo = self.client.getRepository(repoId)

        query = SQLQuery(repo)

        condSet = ConditionSet()

        for cond in conditions:

            kw = cond["keyword"].encode("utf-8")

            if cond["attr"] == "File name":
                query.addCondition(
                    SQLConditionName(kw, cond["operator"]))

            elif cond["attr"] == "Created by":
                query.addCondition(
                    SQLConditionCreatedBy(kw, cond["operator"]))

            elif cond["attr"] == "Modified by":
                query.addCondition(
                    SQLConditionModifiedBy(kw, cond["operator"]))

            elif cond["attr"] == "Creation date":
                query.addCondition(
                    SQLConditionCreationDate(kw, cond["operator"]))

            elif cond["attr"] == "Modification date":
                query.addCondition(
                    SQLConditionModificationDate(
                        kw, cond["operator"]))

            elif cond["attr"] == "Directory path":
                condSet.add(
                    ConditionDirectoryPath(kw, repo))

            elif cond["attr"] == "Comment":
                condSet.add(
                    ConditionComment(kw, cond["operator"]))

        ret = []

        resultSet = query.query()

        for result in resultSet.getResults():
            if condSet.check(result):
                versions = []
                allversions = result.getAllVersions()
                for result2 in allversions.getResults():
                    prop = self.normalizeProperties(result2.properties)
                    prop["path"] = result.getPaths()[0]
                    versions.append(prop)
                ret.append(versions)

        return status, headers, json.dumps(ret)


    def approval (self, repoId, **kwargs):
        headers = [("Content-type", "application/json"), ('Cache-Control','no-cache')]
        status = Status[200]

        doc = None
        path = None

        repo = self.client.getRepository(repoId)

        q = dict(cgi.parse_qsl(self.environ["QUERY_STRING"]))

        doc = None
        if "path" in kwargs:
            path = kwargs["path"]
            doc = repo.getObjectByPath(path)
        elif "fileId" in kwargs:
            doc = repo.getObject(kwargs["fileId"])
        else:
            raise BadRequestException("Bad Request: url")

        if doc.properties["cmis:baseTypeId"] == "cmis:folder":
            raise BadRequestException("Bad Request: path")

        actions = doc.getAllowableActions()
        if not actions["canCheckOut"]:
            raise PermissionDeniedException()

        pwcDoc = doc.checkout()
        try:
            args = {}

            # bug? Two or more versions progress.
            #if "mimeType" in q:
            #    args["contentType"] = q["mimeType"]

            args = {"major": "false"}
            if "comment" not in q:
                raise BadRequestException("Bad Request: comment")

            args["checkinComment"] = q["comment"]

            doc = pwcDoc.checkin(**args)
        finally:
            if doc.isCheckedOut():
                pwcDoc.delete()


        resultSet = doc.getAllVersions()
        results = resultSet.getResults()
        for result in results:
            if result.properties["cmis:isLatestVersion"]:
                break
        prop = self.normalizeProperties(result.properties)
        prop["path"] = doc.getPaths()[0]

        return status, headers, json.dumps(prop)


    def normalizeProperties (self, props):
        props["cmis:creationDate"] = props["cmis:creationDate"].isoformat()
        props["cmis:lastModificationDate"] = props["cmis:lastModificationDate"].isoformat()
        return props



application = Alfresco()
