<?php
/*
 * D-Case Weaver 1.1.1
 *
 * Copyright (c) 2012-2014 JST DEOS R&D Center
 *
 * Licensed under The MIT License
 */

define("ERROR_PERMISSION_DENIED", 1);
define('ERROR_PARSE_FAILED', 2);

class ParseDCaseFailedException extends Exception {}

define('HTTP400', 'HTTP/1.0 400 Bad Request');
define('HTTP403', 'HTTP/1.0 403 Forbidden');
define('HTTP500', 'HTTP/1.0 500 Internal Server Error');

class DCase
{
  var $id;
  var $nodes = array();

  function DCase ($xpath, $elem) {
    $this->id = $elem->getAttribute("id");

    $elems = $xpath->query("dcase:nodes/dcase:node", $elem);
    foreach ($elems as $e) {
      $this->nodes[] = new DCaseNode($xpath, $e);
    }
  }

  function getId () {
    return $this->id;
  }

  function getNodes () {
    return $this->nodes;
  }
}


class DCaseNode
{
  var $name;
  var $id;
  var $type;
  var $dscript;
  var $dscript_name;
  var $values;

  function DCaseNode ($xpath, $elem) {
    if (!$elem->hasAttribute("name") ||
        !$elem->hasAttribute("id") ||
        !$elem->hasAttribute("type"))
      throw new Exception
        ("Failed to parse D-Case.", ERROR_PARSE_FAILED);

    $this->name = $elem->getAttribute("name");
    $this->id = $elem->getAttribute("id");
    $this->type = $elem->getAttribute("type");

    $elems = $xpath->query("dre:d-script", $elem);
    if (0 < $elems->length) {
      $this->dscript = $elems->item(0);

      $elems = $xpath->query("dre:full-name", $this->dscript);
      if (0 < $elems->length)
        $this->dscript_name = $elems->item(0)->textContent;

      $elems = $xpath->query("dre:values/dre:value", $this->dscript);
      foreach ($elems as $e) {
        if ($elem->hasAttribute("name")) {
          $name = $e->getAttribute("name");
          $value = $e->textContent;
          $this->values[$name] = $value;
        }
      }
    }
  }

  function getId () {
    return $this->id;
  }

  function getType () {
    return $this->type;
  }

  function getName () {
    return $this->name;
  }

  function hasDScript () {
    return $this->dscript != NULL;
  }

  function getDScriptName () {
    return $this->dscript_name;
  }

  function getValues () {
    return $this->values;
  }
}

class DScript
{
  var $monitors = array();
  var $actions = array();

  function DScript ($xpath, $elem) {
    $this->xpath = $xpath;
    $elems = $xpath->query("d-script:monitors/d-script:monitor", $elem);
    foreach ($elems as $e) {
      $this->monitors[] = new DScriptModule($xpath, $e);
    }

    $elems = $xpath->query("d-script:actions/d-script:action", $elem);
    foreach ($elems as $e) {
      $this->actions[] = new DScriptModule($xpath, $e);
    }
  }

  function getModule ($node) {
    $modules = NULL;

    $type = $node->getType();

    if ($type == "Action") {
      $modules = $this->actions;
    } else if ($type == "Monitor") {
      $modules = $this->monitors;
    } else {
      return NULL;
    }

    $dscript_name = $node->getDScriptName();

    foreach ($modules as $module) {
      if ($module->getName() != $dscript_name) continue;
      return $module;
    }

    return NULL;
  }
}

class DScriptModule
{
  var $name;
  var $config_template;
  var $config_path;
  var $config_path_extension = "";
  var $parameters = array();

  function DScriptModule ($xpath, $elem) {
    $elems = $xpath->query("d-script:name", $elem);
    if (0 < $elems->length)
      $this->name = $elems->item(0)->textContent;

    $elems = $xpath->query("d-script:config_template", $elem);
    if (0 < $elems->length)
      $this->config_template = $elems->item(0)->textContent;

    $elems = $xpath->query("d-script:config_path", $elem);
    if (0 < $elems->length) {
      $e = $elems->item(0);
      $this->config_path = $e->textContent;
      if ($e->hasAttribute("extension")) {
        $this->config_path_extension = $e->getAttribute("extension");
      } else {
        $this->config_path_extension = "";
      }
    }

    $elems = $xpath->query("d-script:parameters/d-script:parameter", $elem);
    foreach ($elems as $e) {
      $param = array();

      if ($e->hasAttribute("name"))
        $param["name"] = $e->getAttribute("name");

      if ($e->hasAttribute("type"))
        $param["type"] = $e->getAttribute("type");

      if ($e->hasAttribute("min"))
        $param["min"] = $e->getAttribute("min");

      if ($e->hasAttribute("max"))
        $param["max"] = $e->getAttribute("max");

      $els = $xpath->query("d-script:default", $e);
      if (0 < $els->length)
        $param["default"] = $els->item(0)->textContent;

      $els = $xpath->query("d-script:description", $e);
      if (0 < $els->length)
        $param["description"] = $els->item(0)->textContent;

      if (isset($param["name"]))
        $this->parameters[$param["name"]] = $param;
    }

    if (!$this->name ||
        !$this->config_template ||
        !$this->config_path)
      throw new Exception
        ("Failed to parse D-Script Template.", ERROR_PARSE_FAILED);
  }

  function getName () {
    return $this->name;
  }

  function getConfigTemplate () {
    return $this->config_template;
  }

  function getConfigPath () {
    return $this->config_path;
  }

  function getConfigPathExtension () {
    return $this->config_path_extension;
  }

  function getParameters () {
    return $this->parameters;
  }
}


$parameters = array();

function get_value($name, $default_value = NULL) {
    global $parameters;

    if(array_key_exists($name, $parameters)) {
        return $parameters [$name];
    } elseif($default_value !== NULL) {
        return $default_value;
    }
}

function deploy ($dcase, $node, $module) {
  global $parameters;

  $values = $node->getValues();
  $params = $module->getParameters();

  $parameters["d_case_id_by_editor"] = $dcase->getId();
  $parameters["d_case_node_id_by_editor"] = $node->getId();
  $parameters["d_case_node_name"] = $node->getName();

  foreach ($params as $key => $param) {
    if (isset($values[$key])) {
      $parameters[$key] = $values[$key];
    } else {
      $parameters[$key] = $param["default"];
    }
  }

  $config_file = $module->getConfigPath() . $node->getName()
    . $module->getConfigPathExtension();

  if (!file_exists($module->getConfigPath())) {
    if (!mkdir($module->getConfigPath(), 0755, true)) {
      throw new Exception
        ($module->getConfigPath() . ": Permission denied",
         ERROR_PERMISSION_DENIED);
    }
  }

  $template_file = tempnam(sys_get_temp_dir(), 'tmp');

  ob_start();
  foreach(explode(PHP_EOL, $module->getConfigTemplate()) as $line) {
    print(rtrim($line) . ' ' . PHP_EOL);
  }
  $template = ob_get_contents();
  ob_end_clean();

  if (file_put_contents($template_file, $template) === FALSE) {
    throw new Exception
      ($template_file . ": Permission denied", ERROR_PERMISSION_DENIED);
  }

  ob_start();
  include $template_file;
  $config = ob_get_contents();
  ob_end_clean();

  if (file_put_contents($config_file, $config) === FALSE) {
    throw new Exception
      ($config_file . ": Permission denied", ERROR_PERMISSION_DENIED);
  }

  if (posix_getuid() == 0 || posix_getuid() == fileowner($config_file)) {
    chmod($config_file, 0664);
  }

  unlink($template_file);
}

function main () {
  openlog("myScriptLog", LOG_PID, LOG_LOCAL0);

  try {
    header("Content-Type: text/plain; charset: utf-8;");

    $body = file_get_contents('php://input');

    $doc = new DOMDocument;
    if (!$doc->loadXML($body)) {
      throw new Exception
        ("Error parse XML failed", ERROR_PARSE_FAILED);
    }

    $xpath = new DOMXPath($doc);
    $xpath->registerNamespace("dre","http://www.dependable-os.net/dre");
    $xpath->registerNamespace("d-script","http://www.dependable-os.net/d-script");

    $xpath->registerNamespace("dcase","http://www.dependable-os.net/2010/06/dcase");
    $elem = $xpath->query("/dre:deploy/dcase:dcase")->item(0);
    if (!$elem) {
      $xpath->registerNamespace("dcase","http://www.dependable-os.net/2013/11/dcase");
      $elem = $xpath->query("/dre:deploy/dcase:dcase")->item(0);
      if (!$elem)
        throw new Exception
          ("Error parse XML failed", ERROR_PARSE_FAILED);
    }
    $dcase = new DCase($xpath, $elem);
    $nodes = $dcase->getNodes();

    $elem = $xpath->query("/dre:deploy/d-script:d-script")->item(0);
    $dscript = new DScript($xpath, $elem);

    foreach ($nodes as $node) {
      if (!$node->hasDScript()) continue;

      $module = $dscript->getModule($node);
      if ($module == NULL) continue;

      unset($parameters);

      $ret = deploy ($dcase, $node, $module);
    }
  }
  catch (Exception $e) {
    switch ($e->getCode()) {
    case ERROR_PARSE_FAILED:
      header(HTTP400);
      break;
    case ERROR_PERMISSION_DENIED:
      header(HTTP403);
      break;
    default:
      header(HTTP500);
      break;
    }
    echo $e->getMessage();
  }

  closelog();
}

main();
?>