#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# D-Case Weaver 1.1.1
#
# Copyright (c) 2012-2014 JST DEOS R&D Center
#
# Licensed under The MIT License
#

import cgi
import datetime
import fcntl
import json
import mimetypes
import os
import re
import StringIO
import syslog
import time
import traceback


Status = {
    200: "200 OK",
    201: "201 Created",
    400: "400 Bad Request",
    401: "401 Unauthorized",
    404: "404 Not Found",
    500: "500 Internal Server Error",
    550: "550 Permission denied"
};

DOCUMENT_ROOT = u"/var/lib/dcase-weaver/documents"

def debug (msg):
    if isinstance(msg, unicode):
        msg = msg.encode("utf-8")
    elif not isinstance(msg, str):
        msg = str(msg)

    syslog.syslog(syslog.LOG_DEBUG, msg)

class MyException(Exception):

    """
    Common base class for all exceptions.
    """

    def __init__(self, message=None):
        Exception.__init__(self, "Error %s" % message)
        self.message = message

class BadRequestException(MyException):
    pass


class Condition (object):
    def check (self, path):
        pass

class ConditionFileName (Condition):
    def __init__ (self, filename, op):
        self.re = re.compile(filename)
        self.op = op

    def check (self, path):
        m = self.re.search(os.path.basename(path))
        if self.op == "=":
            return m is not None
        elif self.op == "<>":
            return m is None
        else:
            return False

class ConditionDirectoryPath (Condition):
    def __init__ (self, path):
        path = os.path.normpath("/".join((DOCUMENT_ROOT, path)))
        self.re = None
        if os.path.isdir(path.encode('utf-8')):
            self.re = re.compile(path + "/")

    def check (self, path):
        if self.re is None:
            return False

        return self.re.match(path) is not None

class ConditionModificationDate (Condition):
    def __init__ (self, date, op):
        self.date = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
        self.op = op

    def check (self, path):
        mtime = datetime.datetime.utcfromtimestamp(os.path.getmtime(path.encode('utf-8')))

        if self.op == ">=":
            return mtime >= self.date
        elif self.op == "<=":
            return mtime <= self.date
        elif self.op == "=":
            return mtime == self.date
        elif self.op == "<>":
            return mtime != self.date
        else:
            return False

class ConditionSet (Condition):
    def __init__ (self):
        self.conditions = []

    def check (self, path):
        for condition in self.conditions:
            if not condition.check(path):
                return False
        return True

    def add (self, condition):
        self.conditions.append(condition)


class Documents(object):

    def __init__(self):
        pass

    def __call__(self, environ, start_response):
        headers = [("Content-type", "text/plain")]
        status = output = Status[400]

        try:
            self.environ = environ

            method = environ["REQUEST_METHOD"]
            pathinfo = environ["PATH_INFO"].decode('utf-8')
            query = dict(cgi.parse_qsl(environ["QUERY_STRING"]))

            m = re.match("^/metadata(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                if method == "GET":
                    status, headers, output = self.getMetadata(path)
                else:
                    raise BadRequestException("Bad Request Method")
                return

            m = re.match("^/parents(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                if method == "GET":
                    status, headers, output = self.getParents(path)
                else:
                    raise BadRequestException("Bad Request Method")
                return

            m = re.match("^/children(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                if method == "GET":
                    status, headers, output = self.getChildren(path)
                else:
                    raise BadRequestException("Bad Request Method")
                return

            m = re.match("^/file(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                if method == "GET":
                    status, headers, output = self.getFile(path)
                elif method == "PUT":
                    status, headers, output = self.putFile(path)
                elif method == "POST":
                    if "name" not in query:
                        raise BadRequestException("Bad Request: No name")
                    name = query["name"].decode('utf-8')
                    status, headers, output = self.postFile(path, name)
                else:
                    raise BadRequestException("Bad Request Method")
                return

            m = re.match("^/delete(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                status, headers, output = self.delete(path)
                return

            m = re.match("^/mkdir(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                if "name" not in query:
                    raise BadRequestException("Bad Request: No name")
                if path != "/":
                    path += "/"
                path += query["name"].decode('utf-8')
                status, headers, output = self.mkdir(path)
                return

            m = re.match("^/download(/.*)$", pathinfo)
            if m is not None:
                path = m.group(1)
                status, headers, output = self.download(path)
                return

            m = re.match("^/search$", pathinfo)
            if m is not None:
                status, headers, output = self.search(query)
                return

            raise BadRequestException("Bad command")

        except BadRequestException as e:
            status = Status[400]
            output = str(e)

        except OSError as e:
            status = Status[404]
            output = str(e)

        except Exception as e:
            status = Status[500]
            output = str(e)
            debug(traceback.format_exc())

        finally:
            start_response(status, headers)
            return output


    def createMetadata (self, path):
        fullpath = DOCUMENT_ROOT + path

        meta = {}
        meta = {
            "path": path,
            "name": os.path.basename(path)}

        mimetypes.init()
        mimetypes.add_type("application/xml", ".dcase")
        mime = mimetypes.guess_type(fullpath)[0]
        if mime is None:
            mime = "application/octet-stream"
        meta["mime"] = mime

        stat = os.stat(fullpath.encode('utf-8'))
        meta["size"] = stat.st_size
        meta["ctime"] = stat.st_ctime
        meta["mtime"] = stat.st_mtime
        meta["atime"] = stat.st_atime

        meta["isdir"] = os.path.isdir(fullpath.encode('utf-8'))

        return meta;

    def getMetadata (self, path):
        headers = [("Content-type", "application/json")]

        fullpath = DOCUMENT_ROOT + path

        meta = self.createMetadata(path)

        if path == "/":
            path = u""

        if os.path.isdir(fullpath.encode('utf-8')):
            children = []
            for name in os.listdir(fullpath.encode('utf-8')):
                name = name.decode('utf-8')
                children.append(self.createMetadata("%s/%s" % (path, name)))
            meta["children"] = children

        return Status[200], headers, json.dumps(meta)


    def getParents (self, path):
        headers = [("Content-type", "application/json")]

        path = os.path.dirname(path)

        fullpath = DOCUMENT_ROOT + path

        meta = self.createMetadata(path)

        return Status[200], headers, json.dumps([meta])


    def getChildren (self, path):
        headers = [("Content-type", "application/json")]

        fullpath = DOCUMENT_ROOT + path

        meta = self.createMetadata(path)

        children = []

        if os.path.isdir(fullpath.encode('utf-8')):
            if path == "/":
                path = u""

            for name in os.listdir(fullpath.encode('utf-8')):
                name = name.decode('utf-8')
                children.append(self.createMetadata("%s/%s" % (path, name)))

        return Status[200], headers, json.dumps(children)


    def getFile(self, path):
        fullpath = DOCUMENT_ROOT + path

        meta = self.createMetadata(path)

        headers = [("Content-type", meta["mime"])]

        data = None
        with open(fullpath.encode('utf-8')) as f:
            fcntl.flock(f.fileno(), fcntl.LOCK_EX)
            try:
                data = f.read()
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)

        return Status[200], headers, data


    def putFile (self, path):
        fullpath = DOCUMENT_ROOT + path

        wsgiinput = self.environ['wsgi.input']
        length = int(self.environ.get('CONTENT_LENGTH', 0))
        data = wsgiinput.read(length)

        with open(fullpath.encode('utf-8'), "w") as f:
            fcntl.flock(f.fileno(), fcntl.LOCK_EX)
            try:
                f.write(data)
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)

        headers = [("Content-type", "application/json")]

        return Status[200], headers, json.dumps(self.createMetadata(path))


    def postFile (self, path, name):
        if path == "/":
            path += name
        else:
            path += "/" + name

        fullpath = DOCUMENT_ROOT + path

        wsgiinput = self.environ['wsgi.input']
        length = int(self.environ.get('CONTENT_LENGTH', 0))
        data = wsgiinput.read(length)

        with open(fullpath.encode('utf-8'), "w") as f:
            fcntl.flock(f.fileno(), fcntl.LOCK_EX)
            try:
                f.write(data)
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)

        headers = [("Content-type", "application/json")]

        return Status[201], headers, json.dumps(self.createMetadata(path))


    def delete (self, path):
        fullpath = DOCUMENT_ROOT + path

        meta = self.createMetadata(path)

        if os.path.isfile(fullpath.encode('utf-8')):
            os.remove(fullpath.encode('utf-8'))
        elif os.path.isdir(fullpath.encode('utf-8')):
            os.rmdir(fullpath.encode('utf-8'))

        headers = [("Content-type", "application/json")]

        return Status[200], headers, json.dumps(meta)


    def mkdir (self, path):
        fullpath = DOCUMENT_ROOT + path

        os.mkdir(fullpath.encode('utf-8'))

        meta = self.createMetadata(path)

        children = []
        for name in os.listdir(fullpath.encode('utf-8')):
            name = name.decode('utf-8')
            children.append(self.createMetadata("%s/%s" % (path, name)))
        meta["children"] = children

        headers = [("Content-type", "application/json")]

        return Status[200], headers, json.dumps(meta)


    def download (self, path):
        fullpath = DOCUMENT_ROOT + path

        meta = self.createMetadata(path)

        data = ""
        with open(fullpath.encode('utf-8'), "r") as f:
            fcntl.flock(f.fileno(), fcntl.LOCK_EX)
            try:
                data = f.read()
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)

        name = meta["name"].encode('utf-8')

        headers = [
            ("Content-type", meta["mime"]),
            ("Content-Disposition", "attachment; filename=" + name),
            ("Content-Length", str(len(data)))]

        return Status[200], headers, data


    def search (self, query):
        def walk(condition):
            for path, dirs, files in os.walk(DOCUMENT_ROOT.encode('utf-8')):
                path = path.decode('utf-8')
                for filename in files:
                    filename = filename.decode('utf-8')
                    if condition.check(os.path.join(path, filename)):
                        yield "/" + os.path.relpath(os.path.join(path, filename), DOCUMENT_ROOT)
        condition = ConditionSet()

        if "conditions" not in query:
            raise BadRequestException("No condition parameter")

        conds = json.loads(query["conditions"])
        for c in conds:
            if c["attr"] == "File name":
                condition.add(ConditionFileName(c["keyword"], c["operator"]))

            elif c["attr"] == "Directory path":
                condition.add(ConditionDirectoryPath(c["keyword"]))

            elif c["attr"] == "Modification date":
                condition.add(ConditionModificationDate(c["keyword"], c["operator"]))

        ret = []

        for path in walk(condition):
            ret.append(self.createMetadata(path))

        headers = [("Content-type", "application/json")]

        return Status[200], headers, json.dumps(ret)


application = Documents()


