#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# D-Case Weaver 1.1.1
#
# Copyright (c) 2012-2014 JST DEOS R&D Center
#
# Licensed under The MIT License
#

import cgi
import os
import sqlite3
import syslog

Status = {
    200: "200 OK",
    201: "201 Created",
    400: "400 Bad Request",
    401: "401 Unauthorized",
    404: "404 Not Found",
    500: "500 Internal Server Error",
    550: "550 Permission denied"
};

DB = '/var/lib/dcase-weaver/node-status/status.sqlite3'

DEBUG = False

def debug (msg):
    if DEBUG:
        syslog.syslog(syslog.LOG_DEBUG, msg)


class StatusDB (object):
    def __init__ (self):
        if not os.path.isdir(os.path.dirname(DB)):
            os.mkdir(os.path.dirname(DB))

    def __enter__ (self):
        self._con = sqlite3.connect(DB)
        self.create_table()


    def __exit__(self, exc_type, exc_value, traceback):
        self._con.close();


    def create_table (self):
        sql = u'''
create table if not exists Status (
dcase_id varchar(64),
node_id varchar(64),
status varchar(64),
message varchar(256),
UNIQUE(dcase_id, node_id)
);
'''

        self._con.execute(sql)


    def insert (self, dcase_id, node_id, status = 'normal', message = ''):
        sql = "insert or replace into Status values (?, ?, ?, ?)"
        self._con.execute(sql, (dcase_id, node_id, status, message))


    def delete (self, dcase_id, node_id = None):
        sql = "delete from Status where dcase_id = ?"
        if node_id:
            sql += " and node_id = ?"
            self._con.execute(sql, (dcase_id, node_id))
        else:
            self._con.execute(sql, (dcase_id, ))


    def select (self, dcase_id, node_id = None):
        sql = u"select node_id, status, message from Status where dcase_id = ?"
        if node_id:
            sql += " and node_id = ?"
            return self._con.execute(sql, (dcase_id, node_id))
        else:
            return self._con.execute(sql, (dcase_id, ))


    def commit (self):
        self._con.commit()


    def rollback (self):
        self._con.rollback()


class NodeStatus (object):
    def set_status (self, dcase_id, node_id, status = 'normal', message = ''):
        if status == 'normal' and message == '':
            self.delete_status(dcase_id, node_id)
        else:
            db = StatusDB()
            with db:
                try:
                    db.insert(dcase_id, node_id, status, message)
                    db.commit()
                except Exception as e:
                    db.rollback()
                    raise e


    def get_status (self, dcase_id, node_id = None):
        ret = []

        db = StatusDB()
        with db:
            for row in db.select(dcase_id, node_id):
                ret.append({'id': row[0], 'status': row[1], 'message': row[2]})

        return ret


    def delete_status (self, dcase_id, node_id = None):
        db = StatusDB()
        with db:
            try:
                db.delete(dcase_id, node_id)
                db.commit()
            except Exception as e:
                db.rollback()
                raise e


def application (environ, start_response):
    dcase_id = environ["PATH_INFO"].decode('utf-8').strip('/')

    query = dict(cgi.parse_qsl(environ["QUERY_STRING"]))
    node_id = query.get('id')
    status = query.get('status', 'normal').decode('utf-8')
    message = query.get('message', '').decode('utf-8')

    if node_id: node_id = node_id.decode('utf-8')

    method = environ["REQUEST_METHOD"]

    out = u""
    http_status = Status[200]
    headers = [("Content-type", "text/plain; charset=UTF-8")]

    ns = NodeStatus()

    if method == "POST" or method == "PUT":
        ns.set_status(dcase_id, node_id, status, message)

    elif method == 'DELETE':
        ns.delete_status(dcase_id, node_id)

    else:
        for stat in ns.get_status(dcase_id, node_id):
            out += u'id="%(id)s" status="%(status)s" message="%(message)s"\n' \
                % stat

    start_response(http_status, headers)

    return out.encode('utf-8')


if __name__ == '__main__':
    pass
